# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

try:
    from urllib.parse import unquote
except ImportError:
    from urllib import unquote

from knack.util import CLIError
from knack.log import get_logger
from azure.cli.core.util import user_confirmation

from ._docker_utils import (
    request_data_from_registry,
    get_access_credentials,
    RegistryException,
    RepoAccessTokenPermission
)

logger = get_logger(__name__)


ORDERBY_PARAMS = {
    'time_asc': 'timeasc',
    'time_desc': 'timedesc'
}
DEFAULT_PAGINATION = 100


def _get_repository_path(repository=None):
    """Return the path for a repository, or list of repositories if repository is empty.
    """
    if repository:
        return '/acr/v1/{}'.format(repository)
    return '/acr/v1/_catalog'


def _get_tag_path(repository, tag=None):
    """Return the path for a tag, or list of tags if tag is empty.
    """
    if tag:
        return '/acr/v1/{}/_tags/{}'.format(repository, tag)
    return '/acr/v1/{}/_tags'.format(repository)


def _get_manifest_path(repository, manifest=None):
    """Return the path for a manifest, or list of manifests if manifest is empty.
    """
    if manifest:
        return '/acr/v1/{}/_manifests/{}'.format(repository, manifest)
    return '/acr/v1/{}/_manifests'.format(repository)


def _get_manifest_digest(login_server, repository, tag, username, password):
    response = request_data_from_registry(
        http_method='get',
        login_server=login_server,
        path=_get_tag_path(repository, tag),
        username=username,
        password=password,
        result_index='tag')[0]

    if 'digest' in response and response['digest']:
        return response['digest']

    raise CLIError("Could not get the manifest digest for image '{}:{}'.".format(repository, tag))


def _obtain_data_from_registry(login_server,
                               path,
                               username,
                               password,
                               result_index,
                               top=None,
                               orderby=None):
    result_list = []
    execute_next_http_call = True

    params = {
        'n': DEFAULT_PAGINATION,
        'orderby': ORDERBY_PARAMS[orderby] if orderby else None
    }

    while execute_next_http_call:
        execute_next_http_call = False

        # Override the default page size if top is provided
        if top is not None:
            params['n'] = DEFAULT_PAGINATION if top > DEFAULT_PAGINATION else top
            top -= params['n']

        result, next_link = request_data_from_registry(
            http_method='get',
            login_server=login_server,
            path=path,
            username=username,
            password=password,
            result_index=result_index,
            params=params)

        if result:
            result_list += result

        if top is not None and top <= 0:
            break

        if next_link:
            # The registry is telling us there's more items in the list,
            # and another call is needed. The link header looks something
            # like `Link: </v2/_catalog?last=hello-world&n=1>; rel="next"`
            # we should follow the next path indicated in the link header
            next_link_path = next_link[(next_link.index('<') + 1):next_link.index('>')]
            tokens = next_link_path.split('?', 2)
            params = {y[0]: unquote(y[1]) for y in (x.split('=', 2) for x in tokens[1].split('&'))}
            execute_next_http_call = True

    return result_list


def acr_repository_list(cmd,
                        registry_name,
                        top=None,
                        resource_group_name=None,  # pylint: disable=unused-argument
                        tenant_suffix=None,
                        username=None,
                        password=None):
    login_server, username, password = get_access_credentials(
        cmd=cmd,
        registry_name=registry_name,
        tenant_suffix=tenant_suffix,
        username=username,
        password=password)

    return _obtain_data_from_registry(
        login_server=login_server,
        path='/v2/_catalog',
        username=username,
        password=password,
        result_index='repositories',
        top=top)


def acr_repository_show_tags(cmd,
                             registry_name,
                             repository,
                             top=None,
                             orderby=None,
                             resource_group_name=None,  # pylint: disable=unused-argument
                             tenant_suffix=None,
                             username=None,
                             password=None,
                             detail=False):
    login_server, username, password = get_access_credentials(
        cmd=cmd,
        registry_name=registry_name,
        tenant_suffix=tenant_suffix,
        username=username,
        password=password,
        repository=repository,
        permission=RepoAccessTokenPermission.METADATA_READ.value)

    try:
        raw_result = _obtain_data_from_registry(
            login_server=login_server,
            path=_get_tag_path(repository),
            username=username,
            password=password,
            result_index='tags',
            top=top,
            orderby=orderby)
    except RegistryException as e:
        # Check for Classic registry
        if e.status_code == 405:
            if detail:
                logger.warning("The specified --detail is ignored as it is only supported for managed registries.")
            if top:
                logger.warning("The specified --top is ignored as it is only supported for managed registries.")
            if orderby:
                logger.warning("The specified --orderby is ignored as it is only supported for managed registries.")
            return _obtain_data_from_registry(
                login_server=login_server,
                path='/v2/{}/tags/list'.format(repository),
                username=username,
                password=password,
                result_index='tags')
        raise

    # For backward compatibility, convert the results to the old schema
    if not detail:
        return [item['name'] for item in raw_result]

    return raw_result


# This command lists manifest metadata; deprecated for manifest metadata list
def acr_repository_show_manifests(cmd,
                                  registry_name,
                                  repository,
                                  top=None,
                                  orderby=None,
                                  resource_group_name=None,  # pylint: disable=unused-argument
                                  tenant_suffix=None,
                                  username=None,
                                  password=None,
                                  detail=False):
    login_server, username, password = get_access_credentials(
        cmd=cmd,
        registry_name=registry_name,
        tenant_suffix=tenant_suffix,
        username=username,
        password=password,
        repository=repository,
        permission=RepoAccessTokenPermission.METADATA_READ.value)

    raw_result = _obtain_data_from_registry(
        login_server=login_server,
        path=_get_manifest_path(repository),
        username=username,
        password=password,
        result_index='manifests',
        top=top,
        orderby=orderby)

    # For backward compatibility, convert the results to the old schema
    if not detail:
        return [{
            'digest': item['digest'] if 'digest' in item else '',
            'tags': item['tags'] if 'tags' in item else [],
            'timestamp': item['lastUpdateTime'] if 'lastUpdateTime' in item else ''
        } for item in raw_result]

    return raw_result


def acr_repository_show(cmd,
                        registry_name,
                        repository=None,
                        image=None,
                        resource_group_name=None,  # pylint: disable=unused-argument
                        tenant_suffix=None,
                        username=None,
                        password=None):
    return _acr_repository_attributes_helper(
        cmd=cmd,
        registry_name=registry_name,
        http_method='get',
        json_payload=None,
        permission=RepoAccessTokenPermission.METADATA_READ.value,
        repository=repository,
        image=image,
        tenant_suffix=tenant_suffix,
        username=username,
        password=password)


def acr_repository_update(cmd,
                          registry_name,
                          repository=None,
                          image=None,
                          resource_group_name=None,  # pylint: disable=unused-argument
                          tenant_suffix=None,
                          username=None,
                          password=None,
                          delete_enabled=None,
                          list_enabled=None,
                          read_enabled=None,
                          write_enabled=None):
    json_payload = {}

    if delete_enabled is not None:
        json_payload.update({
            'deleteEnabled': delete_enabled
        })
    if list_enabled is not None:
        json_payload.update({
            'listEnabled': list_enabled
        })
    if read_enabled is not None:
        json_payload.update({
            'readEnabled': read_enabled
        })
    if write_enabled is not None:
        json_payload.update({
            'writeEnabled': write_enabled
        })

    permission = RepoAccessTokenPermission.META_WRITE_META_READ.value if json_payload \
        else RepoAccessTokenPermission.METADATA_READ.value
    return _acr_repository_attributes_helper(
        cmd=cmd,
        registry_name=registry_name,
        http_method='patch' if json_payload else 'get',
        json_payload=json_payload,
        permission=permission,
        repository=repository,
        image=image,
        tenant_suffix=tenant_suffix,
        username=username,
        password=password)


def _acr_repository_attributes_helper(cmd,
                                      registry_name,
                                      http_method,
                                      json_payload,
                                      permission,
                                      repository=None,
                                      image=None,
                                      tenant_suffix=None,
                                      username=None,
                                      password=None):
    _validate_parameters(repository, image)

    if image:
        # If --image is specified, repository must be empty.
        repository, tag, manifest = _parse_image_name(image, allow_digest=True)
    else:
        # This is a request on repository
        tag, manifest = None, None

    login_server, username, password = get_access_credentials(
        cmd=cmd,
        registry_name=registry_name,
        tenant_suffix=tenant_suffix,
        username=username,
        password=password,
        repository=repository,
        permission=permission)

    if tag:
        path = _get_tag_path(repository, tag)
        result_index = 'tag'
    elif manifest:
        path = _get_manifest_path(repository, manifest)
        result_index = 'manifest'
    else:
        path = _get_repository_path(repository)
        result_index = None

    return request_data_from_registry(
        http_method=http_method,
        login_server=login_server,
        path=path,
        username=username,
        password=password,
        result_index=result_index,
        json_payload=json_payload)[0]


def acr_repository_untag(cmd,
                         registry_name,
                         image,
                         resource_group_name=None,  # pylint: disable=unused-argument
                         tenant_suffix=None,
                         username=None,
                         password=None):
    repository, tag, _ = _parse_image_name(image)

    login_server, username, password = get_access_credentials(
        cmd=cmd,
        registry_name=registry_name,
        tenant_suffix=tenant_suffix,
        username=username,
        password=password,
        repository=repository,
        permission=RepoAccessTokenPermission.DELETE.value)

    return request_data_from_registry(
        http_method='delete',
        login_server=login_server,
        path=_get_tag_path(repository, tag),
        username=username,
        password=password)[0]


def acr_repository_delete(cmd,
                          registry_name,
                          repository=None,
                          image=None,
                          resource_group_name=None,  # pylint: disable=unused-argument
                          tenant_suffix=None,
                          username=None,
                          password=None,
                          yes=False):
    _validate_parameters(repository, image)

    if image:
        # If --image is specified, repository must be empty.
        repository, tag, manifest = _parse_image_name(image, allow_digest=True)
    else:
        # This is a request on repository
        tag, manifest = None, None

    login_server, username, password = get_access_credentials(
        cmd=cmd,
        registry_name=registry_name,
        tenant_suffix=tenant_suffix,
        username=username,
        password=password,
        repository=repository,
        permission=RepoAccessTokenPermission.DELETE_META_READ.value)

    if tag or manifest:
        manifest = _delete_manifest_confirmation(
            login_server=login_server,
            username=username,
            password=password,
            repository=repository,
            tag=tag,
            manifest=manifest,
            yes=yes)
        path = '/v2/{}/manifests/{}'.format(repository, manifest)
    else:
        user_confirmation("Are you sure you want to delete the repository '{}' "
                          "and all images under it?".format(repository), yes)
        path = _get_repository_path(repository)

    return request_data_from_registry(
        http_method='delete',
        login_server=login_server,
        path=path,
        username=username,
        password=password)[0]


def _validate_parameters(repository, image):
    if bool(repository) == bool(image):
        raise CLIError('Usage error: --image IMAGE | --repository REPOSITORY')


def _parse_image_name(image, allow_digest=False):
    if allow_digest and '@' in image:
        # This is probably an image name by manifest digest
        tokens = image.split('@')
        if len(tokens) == 2:
            return tokens[0], None, tokens[1]

    if ':' in image:
        # This is probably an image name by tag
        tokens = image.split(':')
        if len(tokens) == 2:
            return tokens[0], tokens[1], None
    else:
        # This is probably an image with implicit latest tag
        return image, 'latest', None

    if allow_digest:
        raise CLIError("The name of the image to delete may include a tag in the"
                       " format 'name:tag' or digest in the format 'name@digest'.")
    raise CLIError("The name of the image may include a tag in the format 'name:tag'.")


def _delete_manifest_confirmation(login_server,
                                  username,
                                  password,
                                  repository,
                                  tag,
                                  manifest,
                                  yes):
    # Always query manifest if it is empty
    manifest = manifest or _get_manifest_digest(
        login_server=login_server,
        repository=repository,
        tag=tag,
        username=username,
        password=password)

    tags = _obtain_data_from_registry(
        login_server=login_server,
        path=_get_tag_path(repository),
        username=username,
        password=password,
        result_index='tags')

    filter_by_manifest = [x['name'] for x in tags if manifest == x['digest']]
    message = "This operation will delete the manifest '{}'".format(manifest)
    if filter_by_manifest:
        images = ", ".join(["'{}:{}'".format(repository, str(x)) for x in filter_by_manifest])
        message += " and all the following images: {}".format(images)

    if yes:
        logger.warning(message)
        return manifest

    user_confirmation("{}.\nAre you sure you want to continue?".format(message))

    return manifest


def get_image_digest(cmd, registry_name, image):
    repository, tag, manifest = _parse_image_name(image, allow_digest=True)

    if manifest:
        return repository, tag, manifest

    # If we don't have manifest yet, try to get it from tag.
    login_server, username, password = get_access_credentials(
        cmd=cmd,
        registry_name=registry_name,
        repository=repository,
        permission=RepoAccessTokenPermission.METADATA_READ.value)

    manifest = _get_manifest_digest(
        login_server=login_server,
        repository=repository,
        tag=tag,
        username=username,
        password=password)

    return repository, tag, manifest



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
# pylint: disable=unused-argument, line-too-long
from azure.cli.core.util import sdk_no_wait, CLIError
from azure.mgmt.synapse.models import Workspace, WorkspacePatchInfo, ManagedIdentity, \
    DataLakeStorageAccountDetails, WorkspaceKeyDetails, CustomerManagedKeyDetails, EncryptionDetails, ManagedVirtualNetworkSettings, \
    ManagedIdentitySqlControlSettingsModelPropertiesGrantSqlControlToManagedIdentity, IpFirewallRuleInfo, Key, ManagedIdentitySqlControlSettingsModel, WorkspaceRepositoryConfiguration
from azure.mgmt.cdn.models import CheckNameAvailabilityInput


# Synapse workspace
def list_workspaces(cmd, client, resource_group_name=None):
    return client.list_by_resource_group(
        resource_group_name=resource_group_name) if resource_group_name else client.list()


# pylint: disable=too-many-locals
def create_workspace(cmd, client, resource_group_name, workspace_name, storage_account, file_system,
                     sql_admin_login_user, sql_admin_login_password, location=None, key_name="default", key_identifier=None, enable_managed_virtual_network=None,
                     allowed_aad_tenant_ids=None, prevent_data_exfiltration=None, tags=None, repository_type=None, host_name=None, account_name=None,
                     collaboration_branch=None, repository_name=None, root_folder='/', project_name=None, tenant_id=None, no_wait=False):
    identity_type = "SystemAssigned"
    identity = ManagedIdentity(type=identity_type)
    account_url = "https://{}.dfs.{}".format(storage_account, cmd.cli_ctx.cloud.suffixes.storage_endpoint)
    default_data_lake_storage = DataLakeStorageAccountDetails(account_url=account_url, filesystem=file_system)
    encryption = None
    managed_virtual_network_settings = None
    tenant_ids_list = None
    workspace_repository_configuration = None
    if key_identifier is not None:
        workspace_key_detail = WorkspaceKeyDetails(name=key_name, key_vault_url=key_identifier)
        encryption = EncryptionDetails(cmk=CustomerManagedKeyDetails(key=workspace_key_detail))

    if [''] == allowed_aad_tenant_ids:
        tenant_ids_list = []
    else:
        tenant_ids_list = allowed_aad_tenant_ids

    if enable_managed_virtual_network:
        if prevent_data_exfiltration:
            managed_virtual_network_settings = ManagedVirtualNetworkSettings(prevent_data_exfiltration=True, allowed_aad_tenant_ids_for_linking=tenant_ids_list)
        else:
            managed_virtual_network_settings = ManagedVirtualNetworkSettings(prevent_data_exfiltration=False)

    if repository_type:
        if repository_type == 'AzureDevOpsGit':
            repository_type = 'WorkspaceVSTSConfiguration'
        else:
            repository_type = 'WorkspaceGitHubConfiguration'
        if repository_type == 'WorkspaceVSTSConfiguration' and tenant_id is None:
            from ..util import get_tenant_id
            tenant_id = get_tenant_id()
        if repository_type == 'WorkspaceVSTSConfiguration' and project_name is None:
            from azure.cli.core.azclierror import RequiredArgumentMissingError
            err_msg = 'project_name argument is missing'
            recommendation = 'provide a project name by --project-name'
            raise RequiredArgumentMissingError(err_msg, recommendation)
        workspace_repository_configuration = WorkspaceRepositoryConfiguration(type=repository_type,
                                                                              host_name=host_name,
                                                                              account_name=account_name,
                                                                              project_name=project_name,
                                                                              repository_name=repository_name,
                                                                              collaboration_branch=collaboration_branch,
                                                                              root_folder=root_folder,
                                                                              tenant_id=tenant_id)

    workspace_info = Workspace(
        identity=identity,
        default_data_lake_storage=default_data_lake_storage,
        sql_administrator_login=sql_admin_login_user,
        sql_administrator_login_password=sql_admin_login_password,
        location=location,
        managed_virtual_network="default" if enable_managed_virtual_network is True else None,
        managed_virtual_network_settings=managed_virtual_network_settings,
        encryption=encryption,
        tags=tags,
        workspace_repository_configuration=workspace_repository_configuration
    )
    return sdk_no_wait(no_wait, client.begin_create_or_update, resource_group_name, workspace_name, workspace_info)


# pylint: disable=too-many-locals
def update_workspace(cmd, client, resource_group_name, workspace_name, sql_admin_login_password=None,
                     allowed_aad_tenant_ids=None, tags=None, key_name=None, repository_type=None, host_name=None, account_name=None,
                     collaboration_branch=None, repository_name=None, root_folder=None, project_name=None, tenant_id=None, no_wait=False):
    encryption = None
    tenant_ids_list = None
    workspace_repository_configuration = None

    if key_name:
        workspace_key_detail = WorkspaceKeyDetails(name=key_name)
        encryption = EncryptionDetails(cmk=CustomerManagedKeyDetails(key=workspace_key_detail))

    if allowed_aad_tenant_ids and '' in allowed_aad_tenant_ids:
        tenant_ids_list = []
    else:
        tenant_ids_list = allowed_aad_tenant_ids

    if repository_type:
        if repository_type == 'AzureDevOpsGit':
            repository_type = 'WorkspaceVSTSConfiguration'
        else:
            repository_type = 'WorkspaceGitHubConfiguration'
        if repository_type == 'WorkspaceVSTSConfiguration' and tenant_id is None:
            from ..util import get_tenant_id
            tenant_id = get_tenant_id()
        if repository_type == 'WorkspaceVSTSConfiguration' and project_name is None:
            from azure.cli.core.azclierror import RequiredArgumentMissingError
            err_msg = 'project_name argument is missing'
            recommendation = 'provide a project name by --project-name'
            raise RequiredArgumentMissingError(err_msg, recommendation)
        workspace_repository_configuration = WorkspaceRepositoryConfiguration(type=repository_type,
                                                                              host_name=host_name,
                                                                              account_name=account_name,
                                                                              project_name=project_name,
                                                                              repository_name=repository_name,
                                                                              collaboration_branch=collaboration_branch,
                                                                              root_folder=root_folder,
                                                                              tenant_id=tenant_id)

    updated_vnet_settings = ManagedVirtualNetworkSettings(allowed_aad_tenant_ids_for_linking=tenant_ids_list) if allowed_aad_tenant_ids is not None else None
    workspace_patch_info = WorkspacePatchInfo(tags=tags, sql_admin_login_password=sql_admin_login_password, encryption=encryption, managed_virtual_network_settings=updated_vnet_settings, workspace_repository_configuration=workspace_repository_configuration)
    return sdk_no_wait(no_wait, client.begin_update, resource_group_name, workspace_name, workspace_patch_info)


def get_resource_group_by_workspace_name(cmd, client, workspace_name):
    try:
        return next(workspace for workspace in list_workspaces(cmd, client)
                    if workspace.name == workspace_name).managed_resource_group_name
    except StopIteration:
        raise CLIError('A workspace with name {} does not exist.'.format(workspace_name))


def custom_check_name_availability(cmd, client, name):
    check_name_availability_input = CheckNameAvailabilityInput(name=name, type="Microsoft.Synapse/workspaces")
    return client.check_name_availability(check_name_availability_input)


def create_firewall_rule(cmd, client, resource_group_name, workspace_name, rule_name, start_ip_address, end_ip_address,
                         no_wait=False):
    ip_firewall_rule_info = IpFirewallRuleInfo(name=rule_name, start_ip_address=start_ip_address, end_ip_address=end_ip_address)
    return sdk_no_wait(no_wait, client.begin_create_or_update, resource_group_name, workspace_name, rule_name,
                       ip_firewall_rule_info)


def update_firewall_rule(cmd, client, resource_group_name, workspace_name, rule_name, start_ip_address=None,
                         end_ip_address=None,
                         no_wait=False):
    firewall = client.get(resource_group_name, workspace_name, rule_name)

    start_ip_address = start_ip_address or firewall.start_ip_address
    end_ip_address = end_ip_address or firewall.end_ip_address
    ip_firewall_rule_info = IpFirewallRuleInfo(name=rule_name, start_ip_address=start_ip_address,
                                               end_ip_address=end_ip_address)
    return sdk_no_wait(no_wait, client.begin_create_or_update, resource_group_name, workspace_name, rule_name,
                       ip_firewall_rule_info)


def create_workspace_key(cmd, client, resource_group_name, workspace_name, key_name, key_identifier, no_wait=False):
    key_properties = Key(key_vault_url=key_identifier)
    return sdk_no_wait(no_wait, client.create_or_update, resource_group_name, workspace_name, key_name=key_name, key_properties=key_properties)


def activate_workspace(cmd, client, resource_group_name, workspace_name, key_name, key_identifier, no_wait=False):
    key_properties = Key(is_active_cmk=True, key_vault_url=key_identifier)
    return sdk_no_wait(no_wait, client.create_or_update, resource_group_name, workspace_name, key_name=key_name, key_properties=key_properties)


def grant_sql_access_to_managed_identity(cmd, client, resource_group_name, workspace_name, no_wait=False):
    grant_sql_control_to_managed_identity = ManagedIdentitySqlControlSettingsModelPropertiesGrantSqlControlToManagedIdentity(desired_state="Enabled")
    grant_sql_access_setting = ManagedIdentitySqlControlSettingsModel(grant_sql_control_to_managed_identity=grant_sql_control_to_managed_identity)
    return sdk_no_wait(no_wait, client.begin_create_or_update, resource_group_name, workspace_name, grant_sql_access_setting)


def revoke_sql_access_to_managed_identity(cmd, client, resource_group_name, workspace_name, no_wait=False):
    revoke_sql_control_to_managed_identity = ManagedIdentitySqlControlSettingsModelPropertiesGrantSqlControlToManagedIdentity(
        desired_state="Disabled")
    revoke_sql_access_setting = ManagedIdentitySqlControlSettingsModel(
        grant_sql_control_to_managed_identity=revoke_sql_control_to_managed_identity)
    return sdk_no_wait(no_wait, client.begin_create_or_update, resource_group_name, workspace_name, revoke_sql_access_setting)



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from knack.log import get_logger

logger = get_logger(__name__)


def insert_table_entity(client, table_name, entity, if_exists='fail', timeout=None):
    if if_exists == 'fail':
        return client.insert_entity(table_name, entity, timeout)
    if if_exists == 'merge':
        return client.insert_or_merge_entity(table_name, entity, timeout)
    if if_exists == 'replace':
        return client.insert_or_replace_entity(table_name, entity, timeout)
    from knack.util import CLIError
    raise CLIError("Unrecognized value '{}' for --if-exists".format(if_exists))


def create_table(client, table_name, fail_on_exist=False):
    if fail_on_exist:
        client.create_table(table_name)
    else:
        client.create_table_if_not_exists(table_name)
    return True


def delete_table(client, table_name, fail_not_exist=False):
    if exists(client, table_name):
        client.delete_table(table_name)
        return True
    if fail_not_exist:
        from azure.core.exceptions import ResourceNotFoundError
        raise ResourceNotFoundError("The specified resource does not exist.")
    return False


def list_tables(client, num_results=None, marker=None, show_next_marker=None):
    from ..track2_util import list_generator
    generator = client.list_tables(results_per_page=num_results)
    pages = generator.by_page(continuation_token=marker)
    result = list_generator(pages=pages, num_results=num_results)

    if show_next_marker:
        next_marker = {"nextMarker": pages.continuation_token}
        result.append(next_marker)
    else:
        if pages.continuation_token:
            logger.warning('Next Marker:')
            logger.warning(pages.continuation_token)

    return result


def exists(client, table_name):
    generator = client.query_tables("TableName eq '{}'".format(table_name))
    return list(next(generator.by_page())) != []


# pylint: disable=redefined-builtin
def generate_sas(client, table_name, permission=None, expiry=None, start=None, id=None,
                 ip=None, protocol=None, start_pk=None, start_rk=None, end_pk=None, end_rk=None):
    from azure.data.tables._table_shared_access_signature import generate_table_sas
    return generate_table_sas(client.credential, table_name, permission=permission, expiry=expiry, start=start,
                              policy_id=id, ip_address_or_range=ip, protocol=protocol,
                              start_pk=start_pk, start_rk=start_rk, end_pk=end_pk, end_rk=end_rk)


def insert_entity(client, entity, if_exists='fail'):
    if if_exists == 'fail':
        return client.upsert_entity(entity, mode='merge')
    if if_exists == 'merge' or if_exists == 'replace':
        return client.upsert_entity(entity, mode=if_exists)
    from knack.util import CLIError
    raise CLIError("Unrecognized value '{}' for --if-exists".format(if_exists))


def _update_entity(client, entity, mode, if_match='*'):
    if not if_match or if_match == '*':
        return client.update_entity(entity, mode)
    from azure.core import MatchConditions
    return client.update_entity(entity, mode, etag=if_match, match_condition=MatchConditions.IfNotModified)


def replace_entity(client, entity, if_match='*'):
    return _update_entity(client, entity, mode='replace', if_match=if_match)


def merge_entity(client, entity, if_match='*'):
    return _update_entity(client, entity, mode='merge', if_match=if_match)


def delete_entity(client, partition_key, row_key, if_match='*'):
    if not if_match or if_match == '*':
        return client.delete_entity(partition_key=partition_key, row_key=row_key)
    from azure.core import MatchConditions
    return client.delete_entity(partition_key=partition_key, row_key=row_key,
                                etag=if_match, match_condition=MatchConditions.IfNotModified)


# pylint: disable=redefined-builtin
def query_entity(client, filter=None, select=None, num_results=None, marker=None):
    def _convert_marker_to_continuation_token(mk):
        if not mk:
            return None
        ct = {
            'PartitionKey': mk.get('nextpartitionkey'),
            'RowKey': mk.get('nextrowkey')
        }
        return ct

    def _convert_continuation_token_to_next_marker(ct):
        if not ct:
            return {}
        mk = {
            'nextpartitionkey': ct.get('PartitionKey'),
            'nextrowkey': ct.get('RowKey')
        }
        return mk

    from ..track2_util import list_generator
    generator = client.query_entities(query_filter=filter, results_per_page=num_results, select=select)
    pages = generator.by_page(continuation_token=_convert_marker_to_continuation_token(marker))
    items = list_generator(pages=pages, num_results=num_results)

    result = {
        'items': items,
        'nextMarker': _convert_continuation_token_to_next_marker(pages.continuation_token)
    }
    return result
